from __future__ import annotations
import typing as ty
import re
import glob
import inspect
import shlex
import platform
from pathlib import Path
from copy import copy, deepcopy
import attrs
from fileformats.generic import FileSet, File
from pydra.utils.general import (
    attrs_values,
    get_fields,
    ensure_list,
    position_sort,
)
from pydra.utils.typing import (
    is_fileset_or_union,
    state_array_support,
    is_optional,
    optional_type,
    is_multi_input,
    MultiOutputObj,
    MultiOutputFile,
)
from pydra.compose import base
from pydra.compose.base.field import RequirementSet
from pydra.compose.base.helpers import is_set
from . import field
from .templating import (
    template_update,
    template_update_single,
    argstr_formatting,
    fields_in_formatter,
    parse_format_string,
)

if ty.TYPE_CHECKING:
    from pydra.engine.job import Job

TaskType = ty.TypeVar("TaskType", bound="Task")


@attrs.define(kw_only=True, auto_attribs=False, eq=False, repr=False)
class ShellOutputs(base.Outputs):
    """Output task of a generic shell process."""

    BASE_ATTRS = ["return_code", "stdout", "stderr"]
    RETURN_CODE_HELP = """The process' exit code."""
    STDOUT_HELP = """The standard output stream produced by the command."""
    STDERR_HELP = """The standard error stream produced by the command."""

    return_code: int = field.out(name="return_code", type=int, help=RETURN_CODE_HELP)
    stdout: str = field.out(name="stdout", type=str, help=STDOUT_HELP)
    stderr: str = field.out(name="stderr", type=str, help=STDERR_HELP)

    @classmethod
    def _from_job(cls, job: "Job[Task]") -> ty.Self:
        """Collect the outputs of a shell process from a combination of the provided inputs,
        the objects in the output directory, and the stdout and stderr of the process.

        Parameters
        ----------
        inputs : Task
            The input task of the shell process.
        cache_dir : Path
            The directory where the process was run.
        stdout : str
            The standard output of the process.
        stderr : str
            The standard error of the process.
        return_code : int
            The exit code of the process.

        Returns
        -------
        outputs : Outputs
            The outputs of the shell process
        """
        outputs = super()._from_job(job)
        fld: field.out
        for fld in get_fields(cls):
            if fld.name in ["return_code", "stdout", "stderr"]:
                resolved_value = job.return_values[fld.name]
            # Get the corresponding value from the inputs if it exists, which will be
            # passed through to the outputs, to permit manual overrides
            elif isinstance(fld, field.outarg) and isinstance(
                job.inputs[fld.name], Path
            ):
                resolved_value = job.inputs[fld.name]
            elif is_set(fld.default):
                resolved_value = cls._resolve_default_value(fld, job.cache_dir)
            else:
                resolved_value = cls._resolve_value(fld, job)
            # Set the resolved value
            try:
                setattr(outputs, fld.name, resolved_value)
            except FileNotFoundError:
                if is_optional(fld.type):
                    setattr(outputs, fld.name, None)
                else:
                    raise ValueError(
                        f"file system path(s) provided to mandatory field {fld.name!r}, "
                        f"'{resolved_value}', does not exist, this is likely due to an "
                        f"error in the {job.name!r} job"
                    )
        return outputs

    @classmethod
    def _resolve_default_value(cls, fld: field.out, cache_dir: Path) -> ty.Any:
        """Resolve path and glob expr default values relative to the output dir"""
        default = fld.default
        if fld.type is Path:
            assert isinstance(default, Path)
            if not default.is_absolute():
                default = cache_dir.joinpath(default)
            if "*" not in str(default):
                if default.exists():
                    return default
                else:
                    raise FileNotFoundError(f"file {default} does not exist")
            else:
                all_files = [Path(el) for el in glob(default.expanduser())]
                if len(all_files) > 1:
                    return all_files
                elif len(all_files) == 1:
                    return all_files[0]
                else:
                    raise FileNotFoundError(f"no file matches {default.name}")
        return default

    @classmethod
    def _required_fields_satisfied(cls, fld: field.out, inputs: "Task") -> bool:
        """checking if all fields from the requires and template are set in the input
        if requires is a list of list, checking if at least one list has all elements set
        """

        if not fld.requires:
            return True

        requirements: list[RequirementSet]
        if fld.requires:
            requirements = deepcopy(fld.requires)
        else:
            requirements = [RequirementSet()]

        # if the output has output_file_template field, add in all input fields from
        # the template to requires
        if isinstance(fld, field.outarg) and fld.path_template:
            # if a template is a function it has to be run first with the inputs as the only arg
            if callable(fld.path_template):
                template = fld.path_template(inputs)
            else:
                template = fld.path_template
            inp_fields = re.findall(r"{(\w+)(?:\:[^\}]+)?}", template)
            for req in requirements:
                req += inp_fields

        # Check to see if any of the requirement sets are satisfied
        return any(rs.satisfied(inputs) for rs in requirements)

    @classmethod
    def _resolve_value(
        cls,
        fld: "field.out",
        job: "Job[TaskType]",
    ) -> ty.Any:
        """Collect output file if metadata specified."""

        if not cls._required_fields_satisfied(fld, job.task):
            return None
        if isinstance(fld, field.outarg) and fld.path_template:
            return template_update_single(
                fld,
                task=job.task,
                cache_dir=job.cache_dir,
                spec_type="output",
            )
        assert fld.callable, (
            f"Output field '{fld.name}', does not not contain any of the required fields "
            f'("callable", "output_file_template" or "value"): {fld}.'
        )
        callable_ = fld.callable
        if isinstance(fld.callable, staticmethod):
            # In case callable is defined as a static method,
            # retrieve the function wrapped in the descriptor.
            callable_ = fld.callable.__func__
        call_args = inspect.getfullargspec(callable_)
        call_args_val = {}
        for argnm in call_args.args:
            if argnm == "field":
                call_args_val[argnm] = fld
            elif argnm == "cache_dir":
                call_args_val[argnm] = job.cache_dir
            elif argnm == "executable":
                call_args_val[argnm] = job.task.executable
            elif argnm == "inputs":
                call_args_val[argnm] = job.inputs
            elif argnm == "stdout":
                call_args_val[argnm] = job.return_values["stdout"]
            elif argnm == "stderr":
                call_args_val[argnm] = job.return_values["stderr"]
            elif argnm == "self":
                pass  # If the callable is a class
            else:
                try:
                    call_args_val[argnm] = job.inputs[argnm]
                except KeyError as e:
                    e.add_note(
                        f"arguments of the callable function from {fld.name!r} "
                        f"has to be in inputs or be field or cache_dir, "
                        f"but {argnm!r} is used"
                    )
                    raise
        return callable_(**call_args_val)


ShellOutputsType = ty.TypeVar("OutputType", bound=ShellOutputs)


@state_array_support
def append_args_converter(value: ty.Any) -> list[str]:
    """Convert additional arguments to a list of strings."""
    if isinstance(value, str):
        return shlex.split(value)
    if not isinstance(value, ty.Sequence):
        return [value]
    return list(value)


@attrs.define(kw_only=True, auto_attribs=False, eq=False, repr=False)
class ShellTask(base.Task[ShellOutputsType]):

    _executor_name = "executable"

    BASE_ATTRS = ("append_args",)

    EXECUTABLE_HELP = (
        "the first part of the command, can be a string, "
        "e.g. 'ls', or a list, e.g. ['ls', '-l', 'dirname']"
    )

    append_args: list[str | File] = field.arg(
        name="append_args",
        default=attrs.Factory(list),
        converter=append_args_converter,
        type=list[str | File],
        sep=" ",
        help="Additional free-form arguments to append to the end of the command.",
    )

    RESERVED_FIELD_NAMES = base.Task.RESERVED_FIELD_NAMES + ("cmdline",)

    def _run(self, job: "Job[ShellTask]", rerun: bool = True) -> None:
        """Run the shell command."""
        job.return_values = job.environment.execute(job)

    @property
    def cmdline(self) -> str:
        """The equivalent command line that would be submitted if the job were run on
        the current working directory."""
        # Skip the executable, which can be a multi-part command, e.g. 'docker run'.
        values = attrs_values(self)
        values.update(template_update(self, cache_dir=Path.cwd()))
        cmd_args = self._command_args(values=values)
        cmdline = cmd_args[0]
        for arg in cmd_args[1:]:
            # If there are spaces in the arg, and it is not enclosed by matching
            # quotes, add quotes to escape the space. Not sure if this should
            # be expanded to include other special characters apart from spaces
            if " " in arg:
                cmdline += " '" + arg + "'"
            else:
                cmdline += " " + arg
        return cmdline

    def _command_args(self, values: dict[str, ty.Any]) -> list[str]:
        """Get command line arguments"""
        self._check_resolved()
        self._check_rules()
        # Drop none/empty values and optional path fields that are set to false
        values = copy(values)  # Create a copy so we can drop items from the dictionary
        for fld in get_fields(self):
            fld_value = values[fld.name]
            if fld_value is None or (is_multi_input(fld.type) and fld_value == []):
                del values[fld.name]
            if is_fileset_or_union(fld.type) and type(fld_value) is bool:
                del values[fld.name]
        # Drop special fields that are added separately
        del values["executable"]
        del values["append_args"]
        # Add executable
        pos_args = [
            self._command_shelltask_executable(fld, self.executable),
        ]  # list for (position, command arg)
        positions_provided = [0]
        fields = {f.name: f for f in get_fields(self)}
        for field_name in values:
            pos_val = self._command_pos_args(
                fld=fields[field_name],
                values=values,
                positions_provided=positions_provided,
            )
            if pos_val:
                pos_args.append(pos_val)
        # Sort command and arguments by position
        cmd_args = position_sort(pos_args)
        # pos_args values are each a list of arguments, so concatenate lists after sorting
        command_args = sum(cmd_args, [])
        # Append additional arguments to the end of the command
        command_args += self.append_args
        return command_args

    def _command_shelltask_executable(
        self, fld: field.arg, value: ty.Any
    ) -> tuple[int, ty.Any]:
        """Returning position and value for executable Task input"""
        pos = 0  # executable should be the first el. of the command
        assert value
        return pos, ensure_list(value, tuple2list=True)

    def _command_shelltask_args(
        self, fld: field.arg, value: ty.Any
    ) -> tuple[int, ty.Any]:
        """Returning position and value for args Task input"""
        pos = -1  # assuming that args is the last el. of the command
        if value is None:
            return None
        else:
            return pos, ensure_list(value, tuple2list=True)

    def _command_pos_args(
        self,
        fld: field.arg,
        values: dict[str, ty.Any],
        positions_provided: list[str],
    ) -> tuple[int, ty.Any]:
        """
        Checking all additional input fields, setting pos to None, if position not set.
        Creating a list with additional parts of the command that comes from
        the specific field.

        Parameters
        ----------
        """
        if fld.argstr is None and fld.formatter is None:
            # assuming that input that has no argstr is not used in the command,
            # or a formatter is not provided too.
            return None
        if fld.position is not None:
            if not isinstance(fld.position, int):
                raise Exception(
                    f"position should be an integer, but {fld.position} given"
                )
            # checking if the position is not already used
            if fld.position in positions_provided:
                raise Exception(
                    f"{fld.name} can't have provided position, {fld.position} is already used"
                )

            positions_provided.append(fld.position)

        value = values[fld.name]

        if fld.readonly and type(value) is not bool and value is not attrs.NOTHING:
            raise Exception(f"{fld.name} is read only, the value can't be provided")
        elif value is None and not fld.readonly and fld.formatter is None:
            return None

        cmd_add = []
        # formatter that creates a custom command argument
        # it can take the value of the field, all inputs, or the value of other fields.
        tp = optional_type(fld.type) if is_optional(fld.type) else fld.type
        if fld.formatter:
            call_args = inspect.getfullargspec(fld.formatter)
            call_args_val = {}
            for argnm in call_args.args:
                if argnm == "field":
                    call_args_val[argnm] = fld
                elif argnm == "inputs":
                    call_args_val[argnm] = values
                else:
                    if argnm in values:
                        call_args_val[argnm] = values[argnm]
                    else:
                        raise AttributeError(
                            f"arguments of the formatter function from {fld.name} "
                            f"has to be in inputs or be field, but {argnm} is used"
                        )
            cmd_el_str = fld.formatter(**call_args_val)
            cmd_el_str = cmd_el_str.strip().replace("  ", " ")
            if cmd_el_str != "":
                cmd_add += split_cmd(cmd_el_str)
        elif tp is bool and "{" not in fld.argstr:
            # if value is simply True the original argstr is used,
            # if False, nothing is added to the command.
            if value is True:
                cmd_add.append(fld.argstr)
        elif is_multi_input(tp) or tp is MultiOutputObj or tp is MultiOutputFile:
            # if the field is MultiInputObj, it is used to create a list of arguments
            for val in value or []:
                split_values = copy(values)
                split_values[fld.name] = val
                cmd_add += self._format_arg(fld, split_values)
        else:
            cmd_add += self._format_arg(fld, values)
        return fld.position, cmd_add

    def _format_arg(self, fld: field.arg, values: dict[str, ty.Any]) -> list[str]:
        """Returning arguments used to specify the command args for a single inputs"""
        value = values[fld.name]
        if (
            fld.argstr.endswith("...")
            and isinstance(value, ty.Iterable)
            and not isinstance(value, (str, bytes))
        ):
            argstr = fld.argstr.replace("...", "")
            # if argstr has a more complex form, with "{input_field}"
            if "{" in argstr and "}" in argstr:
                argstr_formatted_l = []
                for val in value:
                    split_values = copy(values)
                    split_values[fld.name] = val
                    argstr_f = argstr_formatting(argstr, split_values)
                    argstr_formatted_l.append(f" {argstr_f}")
                cmd_el_str = fld.sep.join(argstr_formatted_l)
            else:  # argstr has a simple form, e.g. "-f", or "--f"
                cmd_el_str = fld.sep.join([f" {argstr} {val}" for val in value])
        else:
            # in case there are ... when input is not a list
            argstr = fld.argstr.replace("...", "")
            if isinstance(value, ty.Iterable) and not isinstance(value, (str, bytes)):
                cmd_el_str = fld.sep.join([str(val) for val in value])
                value = cmd_el_str
            # if argstr has a more complex form, with "{input_field}"
            if "{" in argstr and "}" in argstr:
                cmd_el_str = argstr.replace(f"{{{fld.name}}}", str(value))
                cmd_el_str = argstr_formatting(cmd_el_str, values)
            else:  # argstr has a simple form, e.g. "-f", or "--f"
                if value:
                    cmd_el_str = f"{argstr} {value}"
                else:
                    cmd_el_str = ""
        return split_cmd(cmd_el_str)

    def _rule_violations(self) -> list[str]:

        errors = super()._rule_violations()
        # if there is a value that has to be updated (e.g. single value from a list)
        # getting all fields that should be formatted, i.e. {field_name}, ...
        fields = get_fields(self)
        available_template_names = [f.name for f in fields] + ["field", "inputs"]
        for fld in fields:
            if fld.argstr:
                if unrecognised := [
                    f
                    for f in parse_format_string(fld.argstr)
                    if f not in available_template_names
                ]:
                    errors.append(
                        f"Unrecognised field names in the argstr of {fld.name} "
                        f"({fld.argstr}): {unrecognised}"
                    )
            if getattr(fld, "path_template", None):
                if unrecognised := [
                    f
                    for f in fields_in_formatter(fld.path_template)
                    if f not in available_template_names
                ]:
                    errors.append(
                        f"Unrecognised field names in the path_template of {fld.name} "
                        f"({fld.path_template}): {unrecognised}"
                    )

        return errors

    DEFAULT_COPY_COLLATION = FileSet.CopyCollation.adjacent


def split_cmd(cmd: str | None):
    """Splits a shell command line into separate arguments respecting quotes

    Parameters
    ----------
    cmd : str
        Command line string or part thereof

    Returns
    -------
    str
        the command line string split into process args
    """
    if cmd is None:
        return []
    # Check whether running on posix or Windows system
    on_posix = platform.system() != "Windows"
    args = shlex.split(cmd, posix=on_posix)
    cmd_args = []
    for arg in args:
        match = re.match("(['\"])(.*)\\1$", arg)
        if match:
            cmd_args.append(match.group(2))
        else:
            cmd_args.append(arg)
    return cmd_args


# Alias ShellTask to Task so we can refer to it by shell.Task
Task = ShellTask
Outputs = ShellOutputs



import typing as ty
import types
from pathlib import Path
from copy import copy
from pydra.utils.typing import TypeParser, is_optional, is_fileset_or_union
import attrs
from .task import Task, Outputs
from pydra.utils.hash import hash_function
from pydra.utils.general import (
    from_list_if_single,
    ensure_list,
    PYDRA_ATTR_METADATA,
)
from pydra.utils.typing import (
    MultiInputObj,
    MultiInputFile,
    MultiOutputObj,
    MultiOutputFile,
    is_lazy,
)
from .field import Field, Arg, Out
from .helpers import sanitize_xor


def build_task_class(
    spec_type: type["Task"],
    out_type: type["Outputs"],
    inputs: dict[str, Arg],
    outputs: dict[str, Out],
    klass: type | None = None,
    name: str | None = None,
    bases: ty.Sequence[type] = (),
    outputs_bases: ty.Sequence[type] = (),
    xor: ty.Sequence[str | None] | ty.Sequence[ty.Sequence[str | None]] = (),
):
    """Create a task class and its outputs class from the
    input and output fields provided to the decorator/function.

    Modifies the class so that its attributes are converted from pydra fields to attrs fields
    and then calls `attrs.define` to create an attrs class (dataclass-like).
    on

    Parameters
    ----------
    task_type : type
        The type of the task to be created
    inputs : dict[str, Arg]
        The input fields of the task
    outputs : dict[str, Out]
        The output fields of the task
    klass : type, optional
        The class to be decorated, by default None
    name : str, optional
        The name of the class, by default
    bases : ty.Sequence[type], optional
        The base classes for the task class, by default ()
    outputs_bases : ty.Sequence[type], optional
        The base classes for the outputs class, by default ()
    xor: Sequence[str | None] | Sequence[Sequence[str | None]], optional
        Names of args that are exclusive mutually exclusive, which must include
        the name of the current field. If this list includes None, then none of the
        fields need to be set.

    Returns
    -------
    klass : type
        The class created using the attrs package
    """
    xor = sanitize_xor(xor)
    spec_type._check_arg_refs(inputs, outputs, xor)

    # Check that the field attributes are valid after all fields have been set
    # (especially the type)
    for inpt in inputs.values():
        attrs.validate(inpt)
    for outpt in outputs.values():
        attrs.validate(outpt)

    if name is None and klass is not None:
        name = klass.__name__
    if reserved_names := [n for n in inputs if n in spec_type.RESERVED_FIELD_NAMES]:
        raise ValueError(
            f"{reserved_names} are reserved and cannot be used for {spec_type} field names"
        )
    outputs_klass = build_outputs_class(out_type, outputs, outputs_bases, name)
    if klass is None:
        if name is None:
            raise ValueError("name must be provided if klass is not")
        bases = tuple(bases)
        # Ensure that Task is a base class
        if not any(issubclass(b, spec_type) for b in bases):
            bases = bases + (spec_type,)
        # If building from a decorated class (as opposed to dynamically from a function
        # or shell-template), add any base classes not already in the bases tuple
        if klass is not None:
            bases += tuple(c for c in klass.__mro__ if c not in bases + (object,))
        # Create a new class with the Task as a base class
        klass = types.new_class(
            name=name,
            bases=bases,
            kwds={},
            exec_body=lambda ns: ns.update({"Outputs": outputs_klass}),
        )
    else:
        # Ensure that the class has it's own annotations dict so we can modify it without
        # messing up other classes
        klass.__annotations__ = copy(klass.__annotations__)
        klass.Outputs = outputs_klass
    # Now that we have saved the attributes in lists to be
    for arg in inputs.values():
        # If an outarg input then the field type should be Path not a FileSet
        attrs_kwargs = _get_attrs_kwargs(arg)
        if isinstance(arg, Out) and is_fileset_or_union(arg.type):
            if getattr(arg, "path_template", False):
                if is_optional(arg.type):
                    field_type = Path | bool | None
                    if arg.mandatory:  # provide default if one is not provided
                        attrs_kwargs["default"] = True if arg.requires else None
                        del attrs_kwargs["factory"]
                else:
                    field_type = Path | bool
                    if arg.mandatory:  # provide default if one is not provided
                        attrs_kwargs["default"] = True  # use the template by default
                        del attrs_kwargs["factory"]
            elif is_optional(arg.type):
                field_type = Path | None
            else:
                field_type = Path
        else:
            field_type = arg.type
        setattr(
            klass,
            arg.name,
            attrs.field(
                converter=make_converter(arg, klass.__name__, field_type),
                validator=make_validator(arg, klass.__name__),
                metadata={PYDRA_ATTR_METADATA: arg},
                on_setattr=attrs.setters.convert,
                **attrs_kwargs,
            ),
        )
        # Store the xor sets for the class
        klass._xor = xor
        klass.__annotations__[arg.name] = field_type

    # Create class using attrs package, will create attributes for all columns and
    # parameters
    attrs_klass = attrs.define(auto_attribs=False, kw_only=True, eq=False, repr=False)(
        klass
    )

    return attrs_klass


def build_outputs_class(
    spec_type: type["Outputs"],
    outputs: dict[str, Out],
    bases: ty.Sequence[type],
    spec_name: str,
) -> type["Outputs"]:
    """Create an outputs class and its outputs class from the
    output fields provided to the decorator/function.

    Creates a new class with attrs fields and then calls `attrs.define` to create an
    attrs class (dataclass-like).

    Parameters
    ----------
    outputs : dict[str, Out]
        The output fields of the task
    bases : ty.Sequence[type], optional
        The base classes for the outputs class, by default ()
    spec_name : str
        The name of the task class the outputs are for

    Returns
    -------
    klass : type
        The class created using the attrs package
    """

    if not any(issubclass(b, spec_type) for b in bases):
        if out_spec_bases := [b for b in bases if issubclass(b, Outputs)]:
            raise ValueError(
                f"Cannot make {spec_type} output definition from {out_spec_bases} bases"
            )
        outputs_bases = bases + (spec_type,)
    if reserved_names := [n for n in outputs if n in spec_type.RESERVED_FIELD_NAMES]:
        raise ValueError(
            f"{reserved_names} are reserved and cannot be used for {spec_type} field names"
        )
    # Add in any fields in base classes that haven't already been converted into attrs
    # fields (e.g. stdout, stderr and return_code)
    for base in outputs_bases:
        base_outputs = {
            n: o
            for n, o in base.__dict__.items()
            if isinstance(o, Out) and n not in outputs
        }
        for name, field in base_outputs.items():
            field.name = name
            field.type = base.__annotations__.get(name, ty.Any)
        outputs.update(base_outputs)
    assert all(o.name == n for n, o in outputs.items())
    outputs_klass = type(
        spec_name + "Outputs",
        tuple(outputs_bases),
        {
            n: attrs.field(
                converter=make_converter(o, f"{spec_name}.Outputs"),
                metadata={PYDRA_ATTR_METADATA: o},
                **_get_attrs_kwargs(o),
            )
            for n, o in outputs.items()
        },
    )
    outputs_klass.__annotations__.update((o.name, o.type) for o in outputs.values())
    outputs_klass = attrs.define(auto_attribs=False, kw_only=True, eq=False)(
        outputs_klass
    )
    return outputs_klass


def make_converter(
    field: Field, interface_name: str, field_type: ty.Type | None = None
) -> ty.Callable[..., ty.Any]:
    """Makes an attrs converter for the field, combining type checking with any explicit
    converters

    Parameters
    ----------
    field : Field
        The field to make the converter for
    interface_name : str
        The name of the interface the field is part of
    field_type : type, optional
        The type of the field, by default None

    Returns
    -------
    converter : callable
        The converter for the field
    """
    if field_type is None:
        field_type = field.type
    checker_label = f"'{field.name}' field of {interface_name} interface"
    type_checker = TypeParser[field_type](
        field_type, label=checker_label, superclass_auto_cast=True
    )
    converters = []
    if field_type in (MultiInputObj, MultiInputFile):
        converters.append(ensure_list)
    elif field_type in (MultiOutputObj, MultiOutputFile):
        converters.append(from_list_if_single)
    if field.converter:
        converters.append(field.converter)
    if converters:
        converters.append(type_checker)
        converter = attrs.converters.pipe(*converters)
    else:
        converter = type_checker
    return converter


def make_validator(field: Field, interface_name: str) -> ty.Callable[..., None] | None:
    """Makes an attrs validator for the field, combining allowed values and any explicit
    validators

    Parameters
    ----------
    field : Field
        The field to make the validator for
    interface_name : str
        The name of the interface the field is part of

    Returns
    -------
    validator : callable
        The validator for the field
    """
    validators = []
    if field.allowed_values:
        validators.append(allowed_values_validator)
    if isinstance(field.validator, ty.Iterable):
        validators.extend(field.validator)
    elif field.validator:
        validators.append(field.validator)
    if len(validators) > 1:
        return validators
    elif validators:
        return validators[0]
    return None


def allowed_values_validator(_, attribute, value):
    """checking if the values is in allowed_values"""
    allowed = attribute.metadata[PYDRA_ATTR_METADATA].allowed_values
    if value is attrs.NOTHING or is_lazy(value):
        pass
    elif value is None and is_optional(attribute.type):
        pass
    elif value not in allowed:
        raise ValueError(
            f"value of {attribute.name} has to be from {allowed}, but {value} provided"
        )


def _get_attrs_kwargs(field: Field) -> dict[str, ty.Any]:
    kwargs = {}
    if not field.mandatory:
        kwargs["default"] = field.default
    # elif is_optional(field.type):
    #     kwargs["default"] = None
    else:
        kwargs["factory"] = nothing_factory
    if field.hash_eq:
        kwargs["eq"] = hash_function
    return kwargs


def nothing_factory():
    return attrs.NOTHING
